#include <DateUtils.hpp>
#include <math.h>
#include <vector.h>

#define VERSION "1.11"

//-------------------------------------------------------------------------
//Schnelle Mathematik-Klassen und Mathematik-Konstanten
//-------------------------------------------------------------------------

const float pi = 3.1415927f;
const float pi2 = 2 * pi;
const float pi_2 = pi/2.0f;
const float pi_180 = pi / 180.0f;
const float w180_pi = 180.0f / pi;
const float vollwinkel = 360.0f;

const int zahlTab1 = 36000;
const int zahlTab2 = 10000;

float degInRad(float winkel)
{
  return winkel * pi_180;
}

float radInDeg(float winkel)
{
  return winkel * w180_pi;
}

class Mathematik
{
  public:
    float tSin[zahlTab1+1];
    float tCos[zahlTab1+1];
    float tTan[zahlTab1+1];
    float tASin[zahlTab2+1];
    float tACos[zahlTab2+1];
    float tATan[zahlTab2+1];

    float vWinkel;
    int rechterWinkelIndex;
    float rechterWinkel;
    float gestreckterWinkel;
    int indexProWinkel;
    float winkelProIndex;

    float skalaASinCos;
    float skalaATan;

  public:
    Mathematik();

    float sinus(int index);
    float cosinus(int index);
    float tangens(int index);

    float aSinus(float wert);
    float aCosinus(float wert);
    float aTangens(float wert);

    int nIndex(int index);               //bringt Index in den Bereich [0;zahlTab1[
    float indexInWinkel(int index);      //Index im Datenfeld wird in Winkel verwandelt
    int winkelInIndex(float winkel);     //Winkel wird in Index im Datenfeld verwandelt
};

Mathematik::Mathematik()
{
  int i;
  float w;

  vWinkel = vollwinkel;
  indexProWinkel = zahlTab1 / vWinkel;
  winkelProIndex = vWinkel / zahlTab1;
  rechterWinkelIndex = zahlTab1 / 4.0f;
  rechterWinkel = vWinkel / 4.0f;
  gestreckterWinkel = vWinkel / 2.0f;

  for(i=0; i<zahlTab1; i++)
  {
    w = degInRad(i * winkelProIndex);
    tSin[i] = sin(w);
    tCos[i] = cos(w);
    tTan[i] = tan(w);
  }

  skalaASinCos = float(zahlTab2);
  skalaATan = zahlTab2 / 100.0f;

  for(i=0; i<zahlTab2; i++)
  {
    tASin[i] = asin(i / skalaASinCos) * w180_pi;
    tACos[i] = acos(i / skalaASinCos) * w180_pi;
    tATan[i] = atan(i / skalaATan) * w180_pi;
  }
}

float Mathematik::sinus(int index)
{
  return tSin[nIndex(index)];
}

float Mathematik::cosinus(int index)
{
  return tCos[nIndex(index)];
}

float Mathematik::tangens(int index)
{
  return tTan[nIndex(index)];
}

float Mathematik::aSinus(float wert)
{
  int index = abs(wert * skalaASinCos);
  return wert >= 0.0f ? tASin[index] : -tASin[index];
}

float Mathematik::aCosinus(float wert)
{
  int index = abs(wert * skalaASinCos);
  return wert >= 0.0f ? tACos[index] : gestreckterWinkel - tACos[index];
}

float Mathematik::aTangens(float wert)
{
  int index = abs(wert * skalaATan);
  if (index < zahlTab2)
  {
    return wert >= 0 ? tATan[index] : -tATan[index];
  }
  else
  {
    return wert >= 0 ? rechterWinkel : -rechterWinkel;
  }
}

int Mathematik::nIndex(int index)
{
  index = index % zahlTab1;
  if (index < 0)
  {
    index = zahlTab1 + index;
  }
  return index;
}

float Mathematik::indexInWinkel(int index)
{
  return float(index) / indexProWinkel;
}

int Mathematik::winkelInIndex(float winkel)
{
  return nIndex(winkel * indexProWinkel);
}

//-------------------------------------------------------------------------
//Vektorklassen und -funktionen
//-------------------------------------------------------------------------
class Vektor2D
{
  public:
    float x,y;
  public:
    Vektor2D();
    Vektor2D(float neuX, float neuY);
};

Vektor2D::Vektor2D()
{
  x = 0;
  y = 0;
}

Vektor2D::Vektor2D(float neuX, float neuY)
{
  x = neuX;
  y = neuY;
}

Vektor2D operator+(Vektor2D v1, Vektor2D v2)
{
  return Vektor2D(v1.x + v2.x, v1.y + v2.y);
}

Vektor2D operator-(Vektor2D v1, Vektor2D v2)
{
  return Vektor2D(v1.x - v2.x, v1.y - v2.y);
}

Vektor2D operator-(Vektor2D v)
{
  return Vektor2D(-v.x, -v.y);
}

Vektor2D operator*(float f, Vektor2D v)
{
  return Vektor2D(f * v.x, f * v.y);
}

Vektor2D operator*(Vektor2D v, float f)
{
  return Vektor2D(f * v.x, f * v.y);
}

Vektor2D operator/(Vektor2D v, float f)
{
  return Vektor2D(v.x / f, v.y / f);
}

float operator*(Vektor2D v1, Vektor2D v2)
{
  return v1.x * v2.x + v1.y * v2.y;
}

bool operator==(Vektor2D v1, Vektor2D v2)
{
  return (v1.x == v2.x) && (v1.y == v2.y);
}

bool operator!=(Vektor2D v1, Vektor2D v2)
{
  return !(v1 == v2);
}

float betrag(Vektor2D v)
{
  return sqrt(v * v);
}

float betragHoch2(Vektor2D v)
{
  return v*v;
}

float winkel(Vektor2D v)
{
  if (v.x == 0)
  {
    if (v.y > 0)
    {
      return pi_2;
    }
    else
    {
      return -pi_2;
    }
  }

  float resultat = atan(v.y / v.x);
  if (v.x < 0)
  {
    resultat = resultat + pi;
  }
  return resultat;
}

Vektor2D einheitsvektor(Vektor2D v)
{
  float b = betrag(v);
  if (b > 0)
  {
    return v/betrag(v);
  }
  else
  {
    return v;
  }
}

Vektor2D einheitsvektor(float winkel)
{
  return Vektor2D(cos(winkel), sin(winkel));
}

Vektor2D senkrecht(Vektor2D v)
{
  return Vektor2D(-v.y, v.x);
}

//-------------------------------------------------------------------------
//Zeichenflaeche 2D
//-------------------------------------------------------------------------
class Zeichenflaeche2D
{
  public:
    TImage *image;
    TCanvas *canvas;
    float xmin, xmax;
    float ymin, ymax;
    float deltaX, deltaY;

  public:
    Zeichenflaeche2D(TImage *neuImage, float neuXMin, float neuXMax,
                     float neuYMin, float neuYMax);
    Zeichenflaeche2D(TImage *neuImage, float neuXMin, float neuXMax,
                     float neuYMin);
    void aktiviereStift(bool neuStatus);
    void setzeStiftModus(TPenMode modus);
    void setzeStiftFarbe(TColor farbe);
    void setzeStiftDicke(int dicke);
    void aktiviereFuellen(bool neuStatus);
    void setzeFuellenFarbe(TColor farbe);

    void loescheZFlaeche();

    int zFlaecheXInImageX(float zx);
    int zFlaecheYInImageY(float zy);
    float imageXInZFlaecheX(int ix);
    float imageYInZFlaecheY(int iy);

    int zFlaecheDXInImageDX(float zdx);
    int zFlaecheDYInImageDY(float zdy);
    float imageDXInZFlaecheDX(int idx);
    float imageDYInZFlaecheDY(int idy);

    void zeichneAchsen(float deltaX, float deltaY);
    void zeichneXAchse(float deltaX);
    void zeichneYAchse(float deltaY);

    void zeichnePunkt(Vektor2D v);
    void zeichneLinie(Vektor2D v1, Vektor2D v2);
    void zeichneLinienzug(Vektor2D v[], int zahlPunkte);
    void zeichneVieleck(Vektor2D v[], int zahlPunkte);
    void zeichneRechteck(Vektor2D v1, Vektor2D v2);
    void zeichneKreis(Vektor2D v, float r);
    void zeichneEllipse(Vektor2D v, float rx, float ry);

    void zeichneGrafik(TGraphic *grafik, Vektor2D v);
    void zeichneGrafik(TGraphic *grafik, Vektor2D v, float groesse);
    void zeichneGrafik(TGraphic *grafik, Vektor2D v,
                       float groesseX, float groesseY);
    void zeichneBildAusListe(TImageList *liste, int index, Vektor2D v);
    void zeichneBildAusListe(TImageList *liste, int index, Vektor2D v,
                             float groesse);
    void zeichneBildAusListe(TImageList *liste, int index, Vektor2D v,
                             float groesseX, float groesseY);

    void zeichneFeder(Vektor2D v1, Vektor2D v2, int zahlSegmente);
    void zeichneSeil(Vektor2D v1, Vektor2D v2, float laenge);

    void positioniereBild(TImage *posImage, Vektor2D v);
    void positioniereBild(TImage *posImage, Vektor2D v, float groesse);
    void positioniereBild(TImage *posImage, Vektor2D v, float groesseX,
                          float groesseY);
};

Zeichenflaeche2D::Zeichenflaeche2D(TImage *neuImage, float neuXMin,
                     float neuXMax, float neuYMin, float neuYMax)
{
  image =neuImage;
  canvas = neuImage->Canvas;

  xmin = neuXMin;
  xmax = neuXMax;
  ymin = neuYMin;
  ymax = neuYMax;

  deltaX = imageDXInZFlaecheDX(1);
  deltaY = imageDYInZFlaecheDY(1);

  loescheZFlaeche();
}

Zeichenflaeche2D::Zeichenflaeche2D(TImage *neuImage,
                     float neuXMin, float neuXMax, float neuYMin)
{
  image =neuImage;
  canvas = neuImage->Canvas;

  xmin = neuXMin;
  xmax = neuXMax;
  ymin = neuYMin;
  ymax = ymin + (neuXMax - neuXMin) * image->Height / image->Width;

  deltaX = imageDXInZFlaecheDX(1);
  deltaY = deltaX;

  loescheZFlaeche();
}

void Zeichenflaeche2D::aktiviereStift(bool status)
{
  if (status)
  {
    canvas->Pen->Style = psSolid;
    setzeStiftModus(pmCopy);
  }
  else
  {
    canvas->Pen->Style = psClear;
  }
}

void Zeichenflaeche2D::setzeStiftModus(TPenMode modus)
{
  canvas->Pen->Mode = modus;
}

void Zeichenflaeche2D::setzeStiftFarbe(TColor farbe)
{
  canvas->Pen->Color = farbe;
}

void Zeichenflaeche2D::setzeStiftDicke(int dicke)
{
  canvas->Pen->Width = dicke;
}

void Zeichenflaeche2D::aktiviereFuellen(bool status)
{
  if (status)
  {
    canvas->Brush->Style = bsSolid;
  }
  else
  {
    canvas->Brush->Style = bsClear;
  }
}

void Zeichenflaeche2D::setzeFuellenFarbe(TColor farbe)
{
  canvas->Brush->Color = farbe;
}

void Zeichenflaeche2D::loescheZFlaeche()
{
  TPenStyle altPenStyle = canvas->Pen->Style;
  TBrushStyle altBrushStyle = canvas->Brush->Style;
  TColor altBrushFarbe = canvas->Brush->Color;

  aktiviereStift(false);
  aktiviereFuellen(true);
  setzeStiftModus(pmCopy);
  setzeFuellenFarbe(clWhite);
  canvas->Rectangle(0, 0, image->Width, image->Height);

  canvas->Pen->Style = altPenStyle;
  canvas->Brush->Style = altBrushStyle;
  setzeFuellenFarbe(altBrushFarbe);
}

int Zeichenflaeche2D::zFlaecheXInImageX(float zx)
{
  return (zx-xmin) / (xmax-xmin) * image->Width;
}

int Zeichenflaeche2D::zFlaecheYInImageY(float zy)
{
  return (ymax-zy) / (ymax-ymin) * image->Height;
}

float Zeichenflaeche2D::imageXInZFlaecheX(int ix)
{
  return xmin + (xmax - xmin) * ix / image->Width;
}

float Zeichenflaeche2D::imageYInZFlaecheY(int iy)
{
  return ymax - (ymax - ymin) * iy / image->Height;
}

int Zeichenflaeche2D::zFlaecheDXInImageDX(float zdx)
{
  return zdx / (xmax-xmin) * image->Width;
}

int Zeichenflaeche2D::zFlaecheDYInImageDY(float zdy)
{
  return zdy / (ymax-ymin) * image->Height;
}

float Zeichenflaeche2D::imageDXInZFlaecheDX(int idx)
{
  return (xmax - xmin) * idx / image->Width;
}

float Zeichenflaeche2D::imageDYInZFlaecheDY(int idy)
{
  return (ymax - ymin) * idy / image->Height;
}

void Zeichenflaeche2D::zeichneXAchse(float deltaX)
{
  setzeStiftModus(pmCopy);
  setzeStiftFarbe(clGray);
  setzeStiftDicke(1);
  int iy = zFlaecheYInImageY(0.0f);
  if (iy >= 0 && iy < image->Height)
  {
    float x;
    int ix;

    canvas->MoveTo(0, iy);
    canvas->LineTo(image->Width, iy);

    x=0.0f;
    while(x < xmax)
    {
      if(x > xmin)
      {
        ix = zFlaecheXInImageX(x);
        canvas->MoveTo(ix, iy-4);
        canvas->LineTo(ix, iy+4);
      }
      x = x + deltaX;
    }

    x = -deltaX;
    while(x > xmin)
    {
      if(x < xmax)
      {
        ix = zFlaecheXInImageX(x);
        canvas->MoveTo(ix, iy-4);
        canvas->LineTo(ix, iy+4);
      }
      x = x - deltaX;
    }

  }
}

void Zeichenflaeche2D::zeichneYAchse(float deltaY)
{
  setzeStiftModus(pmCopy);
  setzeStiftFarbe(clGray);
  setzeStiftDicke(1);
  int ix = zFlaecheXInImageX(0.0f);
  if (ix >= 0 && ix < image->Width)
  {
    float y;
    int iy;
    canvas->MoveTo(ix, 0);
    canvas->LineTo(ix, image->Height);

    y=0.0f;
    while(y < ymax)
    {
      if(y > ymin)
      {
        iy = zFlaecheYInImageY(y);
        canvas->MoveTo(ix-4, iy);
        canvas->LineTo(ix+4, iy);
      }
      y = y + deltaY;
    }

    y = -deltaY;
    while(y > ymin)
    {
      if(y < ymax)
      {
        iy = zFlaecheYInImageY(y);
        canvas->MoveTo(ix-4, iy);
        canvas->LineTo(ix+4, iy);
      }
      y = y - deltaY;
    }

  }
}

void Zeichenflaeche2D::zeichneAchsen(float deltaX, float deltaY)
{
  zeichneXAchse(deltaX);
  zeichneYAchse(deltaY);
}

void Zeichenflaeche2D::zeichnePunkt(Vektor2D v)
{
  int ix = zFlaecheXInImageX(v.x);
  int iy = zFlaecheYInImageY(v.y);

  canvas->Pixels[ix][iy] = clBlack;
}

void Zeichenflaeche2D::zeichneLinie(Vektor2D v1, Vektor2D v2)
{
  int ix1 = zFlaecheXInImageX(v1.x);
  int iy1 = zFlaecheYInImageY(v1.y);
  int ix2 = zFlaecheXInImageX(v2.x);
  int iy2 = zFlaecheYInImageY(v2.y);

  canvas->MoveTo(ix1, iy1);
  canvas->LineTo(ix2, iy2);
}

void Zeichenflaeche2D::zeichneLinienzug(Vektor2D v[], int zahlPunkte)
{
  if (zahlPunkte > 1)
  {
    int i;
    Vektor2D v0 = v[0];
    for (i=1; i<zahlPunkte; i++)
    {
      zeichneLinie(v0, v[i]);
      v0 = v[i];
    }
  }
}

void Zeichenflaeche2D::zeichneVieleck(Vektor2D v[], int zahlPunkte)
{
  if (zahlPunkte > 1)
  {
    int i, ix, iy;
    TPoint *listePunkte;

    listePunkte = new TPoint[zahlPunkte];

    for (i=0; i<zahlPunkte; i++)
    {
      ix = zFlaecheXInImageX(v[i].x);
      iy = zFlaecheYInImageY(v[i].y);
      listePunkte[i] = TPoint(ix, iy);
    }

    canvas->Polygon(listePunkte, zahlPunkte-1);

    delete [] listePunkte;
  }
}

void Zeichenflaeche2D::zeichneKreis(Vektor2D v, float r)
{
  zeichneEllipse(v, r, r);
}

void Zeichenflaeche2D::zeichneEllipse(Vektor2D v, float rx, float ry)
{
  int ix = zFlaecheXInImageX(v.x);
  int iy = zFlaecheYInImageY(v.y);
  int irx = zFlaecheDXInImageDX(rx);
  int iry = zFlaecheDYInImageDY(ry);

  canvas->Ellipse(ix - irx, iy - iry, ix + irx, iy + iry);
}

void Zeichenflaeche2D::zeichneRechteck(Vektor2D v1, Vektor2D v2)
{
  int ix1 = zFlaecheXInImageX(v1.x);
  int iy1 = zFlaecheYInImageY(v1.y);
  int ix2 = zFlaecheXInImageX(v2.x);
  int iy2 = zFlaecheYInImageY(v2.y);

  canvas->Rectangle(ix1, iy1, ix2, iy2);
}

void Zeichenflaeche2D::zeichneFeder(Vektor2D v1, Vektor2D v2,
                                   int zahlSegmente)
{
  Vektor2D abschnitt = (v2 - v1) / (zahlSegmente * 3);
  Vektor2D sAbschnitt = 0.25f * einheitsvektor(senkrecht(abschnitt));
  Vektor2D punkt1 = v1;
  Vektor2D zick, zack, punkt2;

  for (int i=0; i<zahlSegmente; i++)
  {
    zick = punkt1 + abschnitt + sAbschnitt;
    zack = punkt1 + 2 * abschnitt - sAbschnitt;
    punkt2 = punkt1 + 3 * abschnitt;

    zeichneLinie(punkt1, zick);
    zeichneLinie(zick, zack);
    zeichneLinie(zack, punkt2);

    punkt1 = punkt2;
  }
}

void Zeichenflaeche2D::zeichneSeil(Vektor2D v1, Vektor2D v2, float laenge)
{
  float entfernung = betrag(v2 - v1);
  if (entfernung >= 0.999 * laenge)
  {
    zeichneLinie(v1, v2);
  }
  else
  {
    const float teile = 8;
    const float teileMitte = teile/2;
    float teileAbstandVonMitte;
    float durchbiegung = 0.5f * (laenge - entfernung);

    Vektor2D punkt1 = v1;
    Vektor2D punkt2;
    float x0 = v1.x;
    float deltaX = v2.x - x0;
    float y0 = v1.y;
    float deltaY = v2.y - y0;

    for(int i=1; i<=teile; i++)
    {
      teileAbstandVonMitte = i - teileMitte;
      punkt2 = Vektor2D(x0 + deltaX * i / teile,
                        y0 + deltaY * i / teile - durchbiegung *
                        (1.0f -
                         (teileAbstandVonMitte * teileAbstandVonMitte)
                         / (teileMitte * teileMitte))
                       );
      zeichneLinie(punkt1, punkt2);
      punkt1 = punkt2;
    }
  }
}


void Zeichenflaeche2D::zeichneGrafik(TGraphic *grafik, Vektor2D v)
{
  int ix = zFlaecheXInImageX(v.x);
  int iy = zFlaecheYInImageY(v.y);
  int idx = grafik->Width;
  int idy = grafik->Height;

  image->Canvas->Draw(ix - idx / 2, iy - idy / 2, grafik);
}

void Zeichenflaeche2D::zeichneGrafik(TGraphic *grafik, Vektor2D v,
                                     float groesse)
{
  zeichneGrafik(grafik, v, groesse,
                groesse * grafik->Height / grafik->Width);
}

void Zeichenflaeche2D::zeichneGrafik(TGraphic *grafik, Vektor2D v,
                                     float groesseX, float groesseY)
{
  int ix = zFlaecheXInImageX(v.x);
  int iy = zFlaecheYInImageY(v.y);
  int idx = zFlaecheDXInImageDX(groesseX);
  int idy = zFlaecheDYInImageDY(groesseY);

  image->Canvas->StretchDraw(Rect(ix - idx / 2, iy - idy / 2,
                                  ix + idx / 2, iy + idy / 2), grafik);
}

void Zeichenflaeche2D::zeichneBildAusListe(TImageList *liste, int index,
                                           Vektor2D v)
{
  int ix = zFlaecheXInImageX(v.x);
  int iy = zFlaecheYInImageY(v.y);
  int idx = liste->Width;
  int idy = liste->Height;

  liste->Draw(canvas, ix - idx / 2, iy - idy / 2, index, true);
  image->Refresh();
}

void Zeichenflaeche2D::zeichneBildAusListe(TImageList *liste, int index,
                                           Vektor2D v, float groesse)
{
  zeichneBildAusListe(liste, index, v, groesse,
                      groesse * liste->Height / liste->Width);
}

void Zeichenflaeche2D::zeichneBildAusListe(TImageList *liste, int index,
                                           Vektor2D v, float groesseX,
                                           float groesseY)
{
  Graphics::TBitmap *bitmap = new Graphics::TBitmap();

  liste->GetBitmap(index, bitmap);

  int ix = zFlaecheXInImageX(v.x);
  int iy = zFlaecheYInImageY(v.y);
  int idx = zFlaecheDXInImageDX(groesseX);
  int idy = zFlaecheDYInImageDY(groesseY);

  image->Canvas->StretchDraw(Rect(ix - idx / 2, iy - idy / 2,
                                  ix + idx / 2, iy + idy / 2), bitmap);
}

void Zeichenflaeche2D::positioniereBild(TImage *posImage, Vektor2D v)
{
  int ix = zFlaecheXInImageX(v.x);
  int iy = zFlaecheYInImageY(v.y);
  int idx = posImage->Width;
  int idy = posImage->Height;

  if (ix - idx/2 >= 0 && ix + idx/2 <= image->Width &&
      iy - idy/2 >= 0 && iy + idy/2 <= image->Height)
  {
    posImage->Left = image->Left + ix - idx / 2;
    posImage->Top = image->Top + iy - idy / 2;
    posImage->Visible = true;
  }
  else
  {
    posImage->Visible = false;
  }
}

void Zeichenflaeche2D::positioniereBild(TImage *posImage, Vektor2D v,
                                        float groesse)
{
  positioniereBild(posImage, v, groesse,
                   groesse / posImage->Width * posImage->Height);
}

void Zeichenflaeche2D::positioniereBild(TImage *posImage, Vektor2D v,
                                        float groesseX, float groesseY)
{
  int ix = zFlaecheXInImageX(v.x);
  int iy = zFlaecheYInImageY(v.y);
  int idx = zFlaecheDXInImageDX(groesseX);
  int idy = zFlaecheDYInImageDY(groesseY);

  if (ix - idx/2 >= 0 && ix + idx/2 <= image->Width &&
      iy - idy/2 >= 0 && iy + idy/2 <= image->Height)
  {
    posImage->Stretch = false;
    posImage->Width = idx;
    posImage->Height = idy;
    posImage->Left = image->Left + ix - idx / 2;
    posImage->Top = image->Top + iy - idy / 2;
    posImage->Stretch = true;
    posImage->Visible = true;
  }
  else
  {
    posImage->Visible = false;
  }
}

//-------------------------------------------------------------------------
//Physikklassen fr Objekte
//-------------------------------------------------------------------------
const Vektor2D gravBeschl(0, -9.81);
const Vektor2D nullvektor(0,0);


